unit RawRegistry;

// Dedicated to all those, who still read books.

{ ************************************************************************** }
{ * Low level (no WinAPI)                                                  * }
{ * Unit provides READ access only to Win9x/ME registry data               * }
{ * Substitutes standart TRegistry class by TRawRegistry                   * }
{ * Contact me via e-mail: malyarenko(at)pisem.net                         * }
{ ************************************************************************** }

{ -> See standart documentation on TRegistry class for additional info <- }

interface

uses SysUtils, Classes, RegUtils;

{$I regtypes.inc}

const
  MaxSize = 4096;
  ERegistryExceptionText = 'Registry file is corrupted';

type

 ERegistryException = class(Exception)
   constructor Create;
 end;

 TRawRegistry = class (TObject)
  private
    FFileHandle  : TFileStream;

{ General Registry information }

    FBlockCount      : Word;
    FKeyCount        : Integer;
    FFirstBlockOffset: Integer;
    FRootOffset      : Integer;
    FBlockTable      : array of Integer;
    FRootChildOffset : Integer;

{ Current key info }

    FLastKey     : String;   { Last retrieved key name }
    FCurrentKey  : String;
    FCurrentPath : String;
    FHashValue   : Integer;
    FChildOffset : Integer;
    FParentOffset: Integer;
    FNextOffset  : Integer;
    FID          : Word;
    FBlock       : Word;

    FValueCount  : Word;
    FValueOffset : Integer;

{ Current Value Info }

    FDataLen     : Integer;
    FDataType    : Integer;

    function     GetKeyData (const Block, ID: Word): Boolean;
    function     MoveToBlock (const Block, ID: Word): Boolean;

  protected
    procedure    OpenRegistry;
    procedure    OpenKeyHeaderAt(const Offset: Integer);
    function     OpenChildKeyHeader: Boolean;
    function     OpenNextKeyHeader: Boolean;
    function     OpenChildKey(const Key: String): Boolean;
    function     MoveToValue(const ValueName: String): Boolean;
 public
    constructor  Create (const AFileName: String);
    destructor   Destroy; override;

    function     OpenKey(const Path: String): Boolean;
    function     GetKeyInfo(var Value: TRegKeyInfo): Boolean;
    function     HasSubKeys: Boolean;
    procedure    GetKeyNames(Strings: TStrings);
    function     KeyExists(const Path: String): Boolean;

    function     GetDataInfo(const ValueName: String; var Value: TRegDataInfo): Boolean;
    procedure    GetValueNames(Strings: TStrings);
    function     GetDataSize(const ValueName: String): Integer;
    function     ValueExists(const Name: string): Boolean;

    function     ReadBinaryData(const Name: String; var Buffer; BufSize: Integer): Integer;
    function     ReadString(const Name: String): String;
    function     ReadInteger(const Name: String): Integer;
    function     ReadFloat(const Name: String): Double;
    function     ReadCurrency(const Name: String): Currency;
    function     ReadDateTime(const Name: String): TDateTime;
    function     ReadTime(const Name: String): TDateTime;
    function     ReadBool(const Name: String): Boolean;

    property     CurrentPath: String read FCurrentPath;
    property     CurrentKey : String read FCurrentKey;

    property     BlockCount: Word read FBlockCount;
    property     KeyCount  : Integer read FKeyCount;
end;

implementation

{ EInvalidRegistryFormat }

constructor ERegistryException.Create;
begin
  inherited Create(ERegistryExceptionText);
end;

{ TRawRegistry }

constructor TRawRegistry.Create (const AFileName: String);
begin
  inherited Create;
  FFileHandle := TFileStream.Create(AFileName, fmOpenRead or fmShareDenyWrite);
  OpenRegistry;
end;

destructor TRawRegistry.Destroy;
begin
  if FFileHandle <> nil then FFileHandle.Destroy;
  inherited Destroy;
end;

procedure TRawRegistry.OpenRegistry;
var
  CREG: TCREG;
  RGKN: TRGKN;
  RGDB: TRGDB;
  DKE : TDKE;
  I   : Integer;
begin
  if FFileHandle <> nil then
    with FFileHandle do
    begin
      Read(CREG, SizeOf(CREG));
      if CREG.Descriptor <> 'CREG' then raise ERegistryException.Create;
      FFirstBlockOffset  := CREG.Offset;
      FBlockCount        := CREG.BlockCount;
      Read(RGKN, SizeOf(RGKN));
      if RGKN.Descriptor <> 'RGKN' then raise ERegistryException.Create;
      FRootOffset        := RGKN.Offset;
      FKeyCount          := RGKN.KeyCount;
      { Initialize RGDB offsets table }
      SetLength(FBlockTable, FBlockCount);
      Position := FFirstBlockOffset;
      for I := 0 to FBlockCount - 1 do
      begin
        Read(RGDB, SizeOf(RGDB));
        FBlockTable[I] := Position;
        if RGDB.Descriptor <> 'RGDB' then raise ERegistryException.Create;
        Position := Position - $20 + RGDB.Size;
      end;
      { Open Root Key }
      Position := $20 + FRootOffset;
      Read(DKE, SizeOf(DKE));
      FRootChildOffset  := DKE.Child;
      FChildOffset      := FRootChildOffset;
      FCurrentPath      := '\';
    end;
end;

{ ---------------------------------------------------------------------------}
{ 			Key Header Navigation 				     }
{ ---------------------------------------------------------------------------}

procedure TRawRegistry.OpenKeyHeaderAt(const Offset: Integer);
var
  DKE: TDKE;
begin
  if Offset > (FFileHandle.Size - $20) then raise ERegistryException.Create;
  FFileHandle.Position := $20 + Offset;
  FFileHandle.Read(DKE,SizeOf(DKE));
  with DKE do
  begin
    FHashValue    := Hash;
    FChildOffset  := Child;
    FParentOffset := Parent;
    FNextOffset   := Next;
    FID           := ID;
    FBlock        := Block;
  end;
end;

function TRawRegistry.HasSubKeys: Boolean;
begin
  if FChildOffset <> -1 then Result := True
  else Result := False;
end;

function TRawRegistry.OpenChildKeyHeader: Boolean;
begin
  Result := HasSubKeys;
  if Result then OpenKeyHeaderAt(FChildOffset);
end;

function TRawRegistry.OpenNextKeyHeader: Boolean;
begin
  Result := True;
  if FNextOffset <> -1 then OpenKeyHeaderAt(FNextOffset)
  else Result := False;
end;

{ ---------------------------------------------------------------------------}
{ 				RGDB Navigation 			     }
{ ---------------------------------------------------------------------------}

function TRawRegistry.MoveToBlock(const Block,ID: Word): Boolean;
var
  DKH            : TDKH;
  NextBlockOffset: Integer;
begin
  with FFileHandle do
  begin
    if (Block < FBlockCount) then
    begin
      Position := FBlockTable[Block];
      if Block <> FBlockCount - 1 then
        NextBlockOffset := FBlockTable[Block + 1]
      else NextBlockOffset := Size;
      repeat
        Read(DKH, SizeOf(DKH));
        Position := Position + DKH.Size - SizeOf(DKH);
      until (DKH.ID = ID) or (Position > NextBlockOffset);
      if Position > NextBlockOffset then raise ERegistryException.Create;
      Result := True;
      Position := Position - DKH.Size;
    end
    else Result := False;
  end;
end;

function TRawRegistry.GetKeyData(const Block, ID: Word): Boolean;
var
  DKH : TDKH;
  S: array [1..MaxSize] of char;
begin
  Result := MoveToBlock(Block, ID);
  if Result then
  with FFileHandle do
  begin
    Read(DKH,SizeOf(DKH));
    Read(S,DKH.Length);
    FLastKey       := Copy(S,1,DKH.Length);
    FValueCount    := DKH.ValueCount;
    FValueOffset   := Position;
  end;
end;

{ ---------------------------------------------------------------------------}
{ 				Intermediate level                           }
{ ---------------------------------------------------------------------------}

function TRawRegistry.OpenChildKey(const Key: String): Boolean;
var
  Hsh: Integer;

begin
  if Key <> '' then
  begin
    Hsh    := RegHash(Key);
    Result := OpenChildKeyHeader;

    while (FHashValue <> Hsh) and Result do
      Result := OpenNextKeyHeader;
    { Hash collision handling }
    if Result then
    begin
      GetKeyData(FBlock,FID);
      while Result and (AnsiUpperCase(Key) <> AnsiUpperCase(FLastKey)) do
      begin
        Result := OpenNextKeyHeader;
        GetKeyData(FBlock,FID);
      end;
    end;
  end else 
   Result := True;
end;

procedure TRawRegistry.GetKeyNames(Strings: TStrings);
var
  Result: Boolean;
begin
  Strings.Clear;
  Result := OpenChildKeyHeader;
  while Result do
  begin
    GetKeyData(FBlock,FID);
    Strings.Add(FLastKey);
    Result := OpenNextKeyHeader
  end;
  OpenKey(FCurrentPath);
end;

function TRawRegistry.GetKeyInfo(var Value: TRegKeyInfo): Boolean;
var
  I: Integer;
  S: TStringList;
begin
  FillChar(Value,SizeOf(Value),0);
  with Value do
  begin
    Value.NumValues := FValueCount;
    { Walk through Values }
    if FValueCount > 0 then
    begin
      S := TStringList.Create;
      GetValueNames(S);

      for I := 0 to (S.Count - 1) do
      begin
        MaxValueLen := Max(Length(S.Strings[I]),MaxValueLen);
        MaxDataLen  := Max(GetDataSize(S.Strings[I]), MaxDataLen);
      end;
      S.Destroy;
    end;
    { Walk through SubKeys }
    Result := OpenChildKeyHeader;
    while Result do
    begin
      inc(NumSubKeys);
      GetKeyData(FBlock,FID);
      MaxSubKeyLen := Max(Length(FLastKey),MaxSubKeyLen);
      Result := OpenNextKeyHeader;
    end;
    Result := True;
  end;
  OpenKey(FCurrentPath);
end;

function TRawRegistry.OpenKey(const Path: String): Boolean;
var
  P,K,TP: String;
begin
  Result := True;
  if Length(Path) > MaxSize then Result := False
  else
  begin
    TP := Trim(Path);
    P  := TP;
    FChildOffset := FRootChildOffset;

    { Checking for possible 'incorrect' key settings }
    if (P = '\') or (P = '') then Exit;
    if P[1] <> '\' then P := '\' + P;
    if P[Length(P)]='\' then P := Copy(P,Length(P),1);

    while (P <> '') and Result do
    begin
      K := ExtractFirstKey(P);
      Result := OpenChildKey(K);
      Delete(P,1,Length(K)+1);
    end;
    if Result then
    begin
      FCurrentPath := TP;
      FCurrentKey  := ExtractLastKey(TP);
    end;
  end;
end;

function TRawRegistry.KeyExists(const Path: String): Boolean;
var
  OldPath: String;
begin
  OldPath := FCurrentPath;
  Result := OpenKey(Path);
  OpenKey(OldPath);
end;

{ ---------------------------------------------------------------------------}
{ 			Value-based operations				     }
{ ---------------------------------------------------------------------------}

function TRawRegistry.GetDataSize(const ValueName: String): Integer;
var
  Value: TRegDataInfo;
begin
  if GetDataInfo(ValueName,Value) then 
    Result := Value.DataSize else 
    Result := -1;
end;

function TRawRegistry.ValueExists(const Name: string): Boolean;
var
  Value: TRegDataInfo;
begin
  if GetDataInfo(Name,Value) then 
    Result := True else 
    Result := False;
end;

function TRawRegistry.MoveToValue(const ValueName: String): Boolean;
var
  DKV: TDKV;
  I  : Integer;
  S  : array [1..MaxSize] of Char;
  V,UValue : String;
begin
  Result := False;

  if FValueCount <> 0 then
  with FFileHandle do
  begin
    Position := FValueOffset;
    I := 0;
    UValue := AnsiUpperCase(ValueName);
    repeat
      Read(DKV,SizeOf(DKV));
      Read(S,DKV.NameLen);
      V := AnsiUpperCase(Copy(S,1,DKV.NameLen));
      Position := Position + DKV.DataLen;
      inc(I);
    until (I = FValueCount) or (V = UValue);

    if V = UValue then
    begin
      Result := True;
      Position := Position - DKV.DataLen;
      FDataLen  := DKV.DataLen;
      FDataType := DKV.DataType;
    end;
  end;
end;

function TRawRegistry.GetDataInfo(const ValueName: String; var Value: TRegDataInfo): Boolean;
begin
  Result := MoveToValue(ValueName);
  if Result then
  begin
    Value.DataSize := FDataLen;
    case FDataType of
      REG_SZ        : Value.RegData := rdString;
      REG_EXPAND_SZ : Value.RegData := rdExpandString;
      REG_BINARY    : Value.RegData := rdBinary;
      REG_DWORD     : Value.RegData := rdInteger;
    else Value.RegData := rdUnknown;
    end;
  end;
end;

procedure TRawRegistry.GetValueNames(Strings: TStrings);
var
  DKV: TDKV;
  I  : Integer;
  S  : array [1..MaxSize] of Char;
begin
  Strings.Clear;
  if FValueCount <> 0 then
  with FFileHandle do
  begin
    Position := FValueOffset;
    for I := 1 to FValueCount do
    begin
      Read(DKV,SizeOf(DKV));
      Read(S,DKV.NameLen);
      Strings.Add(Copy(S,1,DKV.NameLen));
      Position := Position + DKV.DataLen;
    end;
  end;
end;

function TRawRegistry.ReadBinaryData(const Name: String; var Buffer; BufSize: Integer): Integer;
begin
  Result := FDataLen;
  if MoveToValue(Name) and (FDataType = REG_BINARY) then
    FFileHandle.Read(Buffer,BufSize) else 
    Result := -1;
end;

function TRawRegistry.ReadString(const Name: String): String;
var
  C: array [1..MaxSize] of Char;
begin
  if MoveToValue(Name) and
  ((FDataType = REG_SZ) or (FDataType = REG_EXPAND_SZ)) then
  begin
    FFileHandle.Read(C,FDataLen);
    Result := Copy (C,1,FDataLen);
  end;
end;

function TRawRegistry.ReadInteger(const Name: String): Integer;
var
  I: Integer;
begin
  if MoveToValue(Name) and (FDataType = REG_DWORD) then
    FFileHandle.Read(I,SizeOf(I));
  Result := I;
end;

function TRawRegistry.ReadFloat(const Name: String): Double;
begin
  if MoveToValue(Name) and (FDataType = REG_BINARY) then
    FFileHandle.Read(Result,SizeOf(Result));
end;

function TRawRegistry.ReadCurrency(const Name: String): Currency;
begin
  if MoveToValue(Name) and (FDataType = REG_BINARY) then
    FFileHandle.Read(Result,SizeOf(Result));
end;

function TRawRegistry.ReadDateTime(const Name: String): TDateTime;
begin
  if MoveToValue(Name) and (FDataType = REG_DWORD) then
    FFileHandle.Read(Result,SizeOf(Result));
end;

function TRawRegistry.ReadTime(const Name: String): TDateTime;
begin
  if MoveToValue(Name) and (FDataType = REG_DWORD) then
    FFileHandle.Read(Result,SizeOf(Result));
end;

function TRawRegistry.ReadBool(const Name: String): Boolean;
begin
  if MoveToValue(Name) and (FDataType = REG_DWORD) then
    FFileHandle.Read(Result,SizeOf(Result));
end;

end.