{------------------------------------------------------------------------------}
{                                                                              }
{  Office Assistant                                                            }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

{$I DELPHIAREA.INC}

unit oaActor;

{$R-}

interface

uses
  Windows, Messages, Classes, Controls, Dialogs, Forms, SysUtils, Graphics,
  oaTypes, oaEngine;

type

  EActorError = class(Exception);

  TAssistantActor = class(TPersistent)
  private
    FAnimation: TAssistantEngine;
    FInstance: THandle;
    FPalette: HPalette;
    FTransparentColor: TColor;
    FLibraryName: String;
    FActorInfo: PActorInfo;
    FOnChange: TNotifyEvent;
    ActionCount: array[TActorAction] of Word;
    procedure LoadActorLibrary(LibName: String);
    procedure FreeActorLibrary;
    function InitializeActor: Boolean;
    procedure SetLibraryName(Value: String);
    function GetAnimations(Index: Word): TAnimation;
    function GetAnimationInfo(Index: Word): PAnimationInfo;
    function GetActionAnimationCount(Index: TActorAction): Word;
  public
    constructor Create;
    destructor Destroy; override;
    function Exist: Boolean;
    function ActorName: String;
    function ScreenWidth: Word;
    function ScreenHeight: Word;
    function VersionInfo(ValueName: String): String;
    function AnimationCount: Word;
    function IndexOfAction(Action: TActorAction; Level: Word): Word;
    function IndexOfName(Name: String): Word;
    property Animation: TAssistantEngine read FAnimation;
    property Instance: THandle read FInstance;
    property Palette: HPalette read FPalette;
    property TransparentColor: TColor read FTransparentColor;
    property Animations[Index: Word]: TAnimation read GetAnimations;
    property AnimationInfo[Index: Word]: PAnimationInfo read GetAnimationInfo;
    property ActionAnimationCount[Index: TActorAction]: Word read GetActionAnimationCount;
    property LibraryName: String read FLibraryName write SetLibraryName;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

const
  ActorLibErrors: array[1..2] of String = (
    '%s'#13#10 + 'File is not a valid actor library',
    '%s'#13#10 + 'Actor library file not found');

constructor TAssistantActor.Create;
begin
  inherited Create;
  FAnimation := TAssistantEngine.Create(Self);
  FLibraryName := EmptyStr;
  FActorInfo := nil;
  FInstance := 0;
  FPalette := 0;
  Randomize;
end;

destructor TAssistantActor.Destroy;
begin
  FreeActorLibrary;
  FAnimation.Terminate;
  while FAnimation.Suspended do
    FAnimation.Resume;
  inherited Destroy;
end;

procedure TAssistantActor.LoadActorLibrary(LibName: String);
var
  OldInstance: THandle;
  OldActor: PActorInfo;
begin
  OldInstance := FInstance;
  OldActor := FActorInfo;
  if FileExists(LibName) then
  begin
    FInstance := LoadLibrary(PChar(LibName));
    if (FInstance <> 0) and not InitializeActor then
    begin
      FreeLibrary(FInstance);
      FInstance := 0;
    end;
    if FInstance = 0 then
    begin
      FActorInfo := OldActor;
      FInstance := OldInstance;
      raise EActorError.CreateFmt(ActorLibErrors[1], [LibName]);
    end
    else
    begin
      FAnimation.Clear;
      if OldInstance <> 0 then
        FreeLibrary(OldInstance);
    end;
  end
  else
    raise EActorError.CreateFmt(ActorLibErrors[2], [LibName]);
end;

procedure TAssistantActor.FreeActorLibrary;
begin
  if FInstance <> 0 then
  begin
    FAnimation.Clear;
    if FPalette <> 0 then
      DeleteObject(FPalette);
    FPalette := 0;
    FActorInfo := nil;
    FreeLibrary(FInstance);
    FInstance := 0;
  end;
end;

function TAssistantActor.Exist: boolean;
begin
  Result := FInstance <> 0;
end;

procedure TAssistantActor.SetLibraryName(Value: String);
begin
  if (CompareText(LibraryName, Value) <> 0) then
  begin
    if Value <> EmptyStr then
      LoadActorLibrary(Value)
    else
      FreeActorLibrary;
    FLibraryName := Value;
    if Assigned(FOnChange) then
      FOnChange(Self);
  end;
end;

function TAssistantActor.InitializeActor: Boolean;

  function CreateActorPalette(LogPalette: TMaxLogPalette): HPalette;
  {
  var
    ScreenDC: HDC;
  }
  begin
    {
    ScreenDC := GetDC(0);
    try
      GetSystemPaletteEntries(ScreenDC, 0, 10, LogPalette.palPalEntry[0]);
      GetSystemPaletteEntries(ScreenDC, 246, 10, LogPalette.palPalEntry[246]);
    finally
      ReleaseDC(0, ScreenDC);
    end;
    }
    Result := CreatePalette(PLogPalette(@LogPalette)^);
  end;

var
  GetActorInfo: function: Pointer; stdcall;
  Action: TActorAction;
  ValidLib: Boolean;
  I: Word;
begin
  Result := False;
  if FInstance <> 0 then
  begin
    @GetActorInfo := GetProcAddress(FInstance, 'GetActorInfo');
    if Assigned(GetActorInfo) then
    begin
      try
        FActorInfo := PActorInfo(GetActorInfo);
        ValidLib := Assigned(FActorInfo) and (FActorInfo^.FormatID = 200) and
                   (FActorInfo^.AnimationCount > 0);
      except
        Exit;
      end;
      if ValidLib then
      begin
        for Action := Low(TActorAction) to High(TActorAction) do
          ActionCount[Action] := 0;
        for I := 1 to FActorInfo^.AnimationCount do
          for Action := Low(TActorAction) to High(TActorAction) do
            if Action in FActorInfo^.Animations[I].Actions then
              Inc(ActionCount[Action]);
        if FPalette <> 0 then DeleteObject(FPalette);
        FPalette := CreateActorPalette(FActorInfo^.Palette);
        with FActorInfo^.Palette.palPalEntry[FActorInfo^.TransparentColor] do
          FTransparentColor := TColor(RGB(peRed, peGreen, peBlue));
        Result := True;
      end;
    end;
  end;
end;

function TAssistantActor.VersionInfo(ValueName: String): String;
var
  Trans: PDWord;
  VerInfoSize, ValSize, Dummy: DWord;
  VerInfo, Value: Pointer;
  TransStr, InfoStr: String;
begin
  Result := EmptyStr;
  if FInstance <> 0 then
  begin
    VerInfoSize := GetFileVersioninfoSize(PChar(LibraryName), Dummy);
    if VerInfoSize <> 0 then
    begin
      GetMem(VerInfo, VerInfoSize);
      try
        if GetFileVersionInfo(PChar(LibraryName), Dummy, VerInfoSize, VerInfo) then
        begin
          VerQueryValue(VerInfo, '\VarFileInfo\Translation', Pointer(Trans), valSize);
          TransStr := IntToHex(LoWord(Trans^), 4) + IntToHex(HiWord(Trans^), 4);
          if CompareText(ValueName, 'Language') <> 0 then
          begin
            InfoStr := '\StringFileInfo\' + TransStr + '\' + ValueName;
            if VerQueryValue(VerInfo, PChar(InfoStr), Value, ValSize) then
              Result := StrPas(Value);
          end
          else
          begin
            SetLength(Result, 256);
            VerLanguageName(LoWord(Trans^), PChar(Result), 256);
            SetLength(Result, StrLen(PChar(Result)));
          end;
        end;
      finally
        FreeMem(VerInfo);
      end;
    end;
  end;
end;

function TAssistantActor.ActorName: String;
begin
  Result := EmptyStr;
  if Assigned(FActorInfo) then
    Result := FActorInfo^.ActorName;
end;

function TAssistantActor.AnimationCount: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) then
    Result := FActorInfo^.AnimationCount;
end;

function TAssistantActor.ScreenWidth: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) then
    Result := FActorInfo^.ScreenWidth;
end;

function TAssistantActor.ScreenHeight: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) then
    Result := FActorInfo^.ScreenHeight;
end;

function TAssistantActor.GetActionAnimationCount(Index: TActorAction): Word;
begin
  Result := 0;
  if Assigned(FActorInfo) and (Index in [Low(TActorAction)..High(TActorAction)]) then
    Result := ActionCount[Index];
end;

function TAssistantActor.GetAnimationInfo(Index: Word): PAnimationInfo;
begin
  Result := nil;
  if Assigned(FActorInfo) and (Index >= 1) and (Index <= FActorInfo^.AnimationCount) then
    Result := @(FActorInfo^.Animations[Index]);
end;

function TAssistantActor.GetAnimations(Index: Word): TAnimation;
var
  RetIndex: Word;
begin
  FillChar(Result, SizeOf(TAnimation), 0);
  if Assigned(FActorInfo) and (Index >= 1) and (Index <= FActorInfo^.AnimationCount) then
  begin
    Result.Name := FActorInfo^.Animations[Index].Name;
    Result.Actions := FActorInfo^.Animations[Index].Actions;
    Result.ExitType := FActorInfo^.Animations[Index].ExitType;
    if Result.ExitType = etAnimation then
      RetIndex := FActorInfo^.Animations[Index].ReturnIndex
    else
      RetIndex := 0;
    if (RetIndex >= 1) and (RetIndex <= FActorInfo^.AnimationCount) then
      Result.ReturnName := FActorInfo^.Animations[RetIndex].Name
    else
      Result.ReturnName := EmptyStr;
  end;
end;

function TAssistantActor.IndexOfAction(Action: TActorAction; Level: Word): Word;
var
  FoundLevel, I: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) and (Action <> aaNone) and (ActionCount[Action] > 0) then
  begin
    if Level = 0 then
      Level := Random(ActionCount[Action]) + 1
    else if Level > ActionCount[Action] then
      Level := Level mod ActionCount[Action];
    FoundLevel := 0;
    for I := 1 to FActorInfo^.AnimationCount do
      if Action in FActorInfo^.Animations[I].Actions then
      begin
       Inc(FoundLevel);
       Result := I;
       if FoundLevel >= Level then Break;
      end;
  end;
end;

function TAssistantActor.IndexOfName(Name: String): Word;
var
  I: Word;
begin
  Result := 0;
  if Assigned(FActorInfo) and (Name <> EmptyStr) then
    for I := 1 to FActorInfo^.AnimationCount do
      if CompareText(FActorInfo^.Animations[I].Name, Name) = 0 then
      begin
        Result := I;
        Break;
      end;
end;

end.
