{------------------------------------------------------------------------------}
{                                                                              }
{  Office Assistant                                                            }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

{$I DELPHIAREA.INC}

unit oaDirect;

interface

uses
  Windows, Classes, Graphics, ExtCtrls, oaEngine, oaActor, oaTypes;

type

  TRequestQueue = class;

  TAssistantDirector = class(TPersistent)
  private
    FActor: TAssistantActor;
    FAutoIdle: Boolean;
    FIdleLevel: TIdleLevel;
    FEnabled: Boolean;
    Queue: TRequestQueue;
    IdleTimer: TTimer;
    Returning: Boolean;
    ID: TRequestID;
    IdleID: TRequestID;
    IdlingTime: Boolean;
    IdleLevelTry: Integer;
    FOnPlayStart: TPlayEvent;
    FOnPlayStop: TPlayEvent;
    FOnIdle: TIdleEvent;
    FOnActorChange: TNotifyEvent;
    FOnFrameChange: TNewFrameEvent;
    procedure SetEnabled(Value: Boolean);
    procedure SetAutoIdle(Value: Boolean);
    procedure SetIdleLevel(Value: TIdleLevel);
    procedure ActorChanged(Sender: TObject);
    procedure FrameChanged(Sender: TObject; FrameBitmap: TBitmap);
    procedure AnimationStopped(Sender: TObject);
    procedure IdleTimerExpire(Sender: TObject);
    procedure GoToIdleState;
  protected
    procedure PlayQueuedRequest;
  public
    constructor Create;
    destructor Destroy; override;
    function PlayIndex(Index: Word): TRequestID;
    function PlayName(Name: String): TRequestID;
    function PlayAction(Action: TActorAction): TRequestID;
    procedure Stop(RequestID: TRequestID);
    procedure StopAll;
    function Playing: Boolean;
    function PlayingID: TRequestID;
    function PlayIdling: Boolean;
    function WaitFor(RequestID: TRequestID): Boolean;
    function IsValid(RequestID: TRequestID): Boolean;
    function OrientationOf(Kind: TOrientedAction; Org, Trg: TPoint): TActorAction;
    property Actor: TAssistantActor read FActor;
    property IdleLevel: TIdleLevel read FIdleLevel write SetIdleLevel;
  published
    property AutoIdle: Boolean read FAutoIdle write SetAutoIdle default True;
    property Enabled: Boolean read FEnabled write SetEnabled default True;
    property OnPlayStart: TPlayEvent read FOnPlayStart write FOnPlayStart;
    property OnPlayStop: TPlayEvent read FOnPlayStop write FOnPlayStop;
    property OnIdle: TIdleEvent read FOnIdle write FOnIdle;
    property OnActorChange: TNotifyEvent read FOnActorChange write FOnActorChange;
    property OnFrameChange: TNewFrameEvent read FOnFrameChange write FOnFrameChange;
  end;

  TRequestQueue = class(TList)
  private
    LastID: TRequestID;
    function NewRequest(AnimationIndex: Word): TRequestID;
    procedure RemoveRequest(RequestID: TRequestID);
    function RequestExists(RequestID: TRequestID): Boolean;
    function FetchCurrentTurn: TRequestID;
  end;

implementation

uses
  Forms;

{ TRequestQueue }

function TRequestQueue.NewRequest(AnimationIndex: Word): TRequestID;
begin
  Result := 0;
  if AnimationIndex > 0 then
  begin
    Inc(LastID);
    Result := (LoWord(LastID) shl 16) or AnimationIndex;
    if Add(Pointer(Result)) < 0 then Result := 0;
  end;
end;

procedure TRequestQueue.RemoveRequest(RequestID: TRequestID);
var
  Index: Integer;
begin
  Index := IndexOf(Pointer(RequestID));
  if Index >= 0 then Delete(Index);
end;

function TRequestQueue.RequestExists(RequestID: TRequestID): Boolean;
begin
  Result := IndexOf(Pointer(RequestID)) >= 0;
end;

function TRequestQueue.FetchCurrentTurn: TRequestID;
begin
  Result := 0;
  if Count > 0 then
  begin
    Result := TRequestID(Items[0]);
    Delete(0);
  end;
end;

{ TAssistantDirector }

constructor TAssistantDirector.Create;
begin
  inherited Create;
  Randomize;
  FActor := TAssistantActor.Create;
  FActor.OnChange := ActorChanged;
  FActor.Animation.OnNewFrame := FrameChanged;
  FActor.Animation.OnStop := AnimationStopped;
  IdleTimer := TTimer.Create(nil);
  IdleTimer.Enabled := False;
  IdleTimer.Interval := 3000;
  IdleTimer.OnTimer := IdleTimerExpire;
  Queue := TRequestQueue.Create;
  FAutoIdle := True;
  FEnabled := True;
  IdleLevel := Low(TIdleLevel);
  Returning := False;
  IdlingTime := False;
  ID := 0;
end;

destructor TAssistantDirector.Destroy;
begin
  IdleTimer.Free;
  FActor.Free;
  Queue.Free;
  inherited Destroy;
end;

procedure TAssistantDirector.SetEnabled(Value: Boolean);
begin
  if FEnabled <> Value then
  begin
    FEnabled := Value;
    if not FEnabled then
      IdleTimer.Enabled := False
    else if (ID <> 0) or (Queue.Count > 0) then
      PlayQueuedRequest;
  end;
end;

procedure TAssistantDirector.SetAutoIdle(Value: Boolean);
begin
  if AutoIdle <> Value then
  begin
    FAutoIdle := Value;
    if not AutoIdle then
      IdleLevel := Low(TIdleLevel);
    GoToIdleState;
  end;
end;

procedure TAssistantDirector.SetIdleLevel(Value: TIdleLevel);
var
  M: Integer;
begin
  if (Value in [Low(TIdleLevel)..High(TIdleLevel)]) then
  begin
    FIdleLevel := Value;
    if (IdleLevel = Low(TIdleLevel)) and (ID = IdleID) and
       Actor.Animation.Active and Actor.Animation.HasInfiniteLoop then
      Stop(IdleID);
    M := Ord(High(TIdleLevel)) - Ord(FIdleLevel) + 1;
    IdleLevelTry := 4 * M + Random(10);
  end;
end;

procedure TAssistantDirector.IdleTimerExpire(Sender: TObject);
var
  Handled: Boolean;
  ThisLevel: Integer;
begin
  IdleTimer.Enabled := False;
  IdlingTime := True;
  try
    Handled := False;
    if Assigned(FOnIdle) then
    begin
      ThisLevel := Ord(FIdleLevel) - Ord(Low(TIdleLevel)) + 1;
      try
        FOnIdle(Self, ThisLevel, Handled);
      finally
        FIdleLevel := TIdleLevel(Ord(Low(TIdleLevel)) + ThisLevel - 1);
      end;
    end;
    if not Handled then
    begin
      IdleID := Queue.NewRequest(Actor.IndexOfAction(FIdleLevel, 0));
      if (ID <> 0) or (IdleID > 0) then PlayQueuedRequest;
    end
    else if Queue.Count > 0 then
      IdleID := TRequestID(Queue.Items[0])
    else
      IdleID := ID;
    Dec(IdleLevelTry);
    if IdleLevelTry = 0 then
      IdleLevel := Succ(IdleLevel);
    IdleTimer.Interval := 3000 + Random(1000) - Random(1000);  // 3 +/- 1 sec
  finally
    IdlingTime := False;
    GoToIdleState;
  end;
end;

procedure TAssistantDirector.GoToIdleState;
begin
  IdleTimer.Enabled := AutoIdle and Enabled and not IdlingTime
    and Actor.Exist and not Actor.Animation.Active and (Queue.Count = 0);
end;

procedure TAssistantDirector.ActorChanged(Sender: TObject);
begin
  StopAll;
  if Actor.Exist then
    GoToIdleState;
  if Assigned(FOnActorChange) then
    FOnActorChange(Self);
end;

procedure TAssistantDirector.FrameChanged(Sender: TObject; FrameBitmap: TBitmap);
begin
  if Assigned(FOnFrameChange) then
    FOnFrameChange(Self, FrameBitmap);
end;

procedure TAssistantDirector.PlayQueuedRequest;
begin
  if Enabled and Actor.Exist and not Actor.Animation.Active then
  begin
    IdleTimer.Enabled := False;
    if (ID <> 0) and Actor.Animation.PrepareReturn then
    begin
      Returning := True;
      Actor.Animation.Active := True;
    end
    else if Queue.Count > 0 then
    begin
      ID := Queue.FetchCurrentTurn;
      if Actor.Animation.Prepare(LoWord(ID)) then
        Actor.Animation.Active := True;
    end
    else
      ID := 0;
    if not Returning and (ID <> 0) and (ID <> IdleID) and Assigned(FOnPlayStart) then
      FOnPlayStart(Self, ID);
    GoToIdleState;
  end;
end;

procedure TAssistantDirector.AnimationStopped(Sender: TObject);
var
  TriggerEvent: Boolean;
  StoppedID: TRequestID;
begin
  StoppedID := ID;
  TriggerEvent := not Returning and (StoppedID <> IdleID);
  if StoppedID = IdleID then IdleID := 0;
  if Returning then ID := 0;
  Returning := False;
  try
    if TriggerEvent and Assigned(FOnPlayStop) then
      FOnPlayStop(Self, StoppedID);
  finally
    if Queue.Count > 0 then
      PlayQueuedRequest
    else
      GoToIdleState;
  end;
end;

function TAssistantDirector.PlayIndex(Index: Word): TRequestID;
begin
  Result := 0;
  if (Index > 0) and (Index <= Actor.AnimationCount) then
  begin
    IdleLevel := Low(TIdleLevel);
    if not Returning and (ID <> 0) and (ID = IdleID) then
      Stop(IdleID);
    Result := Queue.NewRequest(Index);
    PlayQueuedRequest;
  end;
end;

function TAssistantDirector.PlayName(Name: String): TRequestID;
begin
  Result := PlayIndex(Actor.IndexOfName(Name));
end;

function TAssistantDirector.PlayAction(Action: TActorAction): TRequestID;
begin
  Result := PlayIndex(Actor.IndexOfAction(Action, 0));
end;

function TAssistantDirector.WaitFor(RequestID: TRequestID): Boolean;
begin
  Result := True;
  if RequestID <> 0 then
  begin
    while (Result = True) and (Queue.RequestExists(RequestID) or
      ((ID = RequestID) and Actor.Animation.Active and not Returning)) do
    begin
      if not Actor.Animation.HasInfiniteLoop then
        repeat
          Application.ProcessMessages
        until not Actor.Animation.Active
      else
        Result := False;
    end;
  end;
end;

procedure TAssistantDirector.Stop(RequestID: TRequestID);
begin
  if RequestID <> 0 then
    if Actor.Animation.Active and (ID = RequestID) then
      Actor.Animation.Active := False
    else
      Queue.RemoveRequest(RequestID);
end;

procedure TAssistantDirector.StopAll;
begin
  Queue.Clear;
  if Actor.Animation.Active then
    Actor.Animation.Active := False;
  ID := 0;
end;

function TAssistantDirector.Playing: Boolean;
begin
  Result := Actor.Animation.Active or (Queue.Count > 0);
end;

function TAssistantDirector.PlayingID: TRequestID;
begin
  Result := ID;
end;

function TAssistantDirector.PlayIdling: Boolean;
begin
  Result := Actor.Animation.Active and (ID <> 0) and (ID = IdleID);
end;

function TAssistantDirector.IsValid(RequestID: TRequestID): Boolean;
begin
  Result := (RequestID <> 0) and
            (((RequestID = ID) and Actor.Animation.Active and not Returning) or
             Queue.RequestExists(RequestID));
end;

function TAssistantDirector.OrientationOf(Kind: TOrientedAction;
  Org, Trg: TPoint): TActorAction;
type
  TDirX = (xNone, Left, Right);
  TDirY = (yNone, Down, Up);
const
  Actions: array[TOrientedAction, TDirX, TDirY] of TActorAction = (
    // oaGesture
    ((aaNone, aaGestureDown, aaGestureUp),
     (aaGestureLeft, aaGestureDownLeft, aaGestureUpLeft),
     (aaGestureRight, aaGestureDownRight, aaGestureUpRight)),
    // oaGlance
    ((aaNone, aaGlanceDown, aaGlanceUp),
     (aaGlanceLeft, aaGlanceDownLeft, aaGlanceUpLeft),
     (aaGlanceRight, aaGlanceDownRight, aaGlanceUpRight)),
    // oaLook
    ((aaNone, aaLookDown, aaLookUp),
     (aaLookLeft, aaLookDownLeft, aaLookUpLeft),
     (aaLookRight, aaLookDownRight, aaLookUpRight)),
    // oaMove
    ((aaNone, aaMoveDown, aaMoveUp),
     (aaMoveLeft, aaMoveDownLeft, aaMoveUpLeft),
     (aaMoveRight, aaMoveDownRight, aaMoveUpRight)));
var
  DirX: TDirX;
  DirY: TDirY;
begin
  if (Abs(Org.x - Trg.x) > (Actor.ScreenWidth div 2)) or (Kind = oaMove) then
    if Org.x > Trg.x then
      DirX := Right
    else
      DirX := Left
  else
    DirX := xNone;
  if (Abs(Org.y - Trg.y) > (Actor.ScreenHeight div 2)) or (Kind = oaMove) then
    if Org.y > Trg.y then
      DirY := Up
    else
      DirY := Down
  else
    DirY := yNone;
  Result := Actions[Kind, DirX, DirY];
  if Actor.IndexOfAction(Result, 0) = 0 then
  begin
    if Abs(Org.x - Trg.x) < Abs(Org.y - Trg.y) then
      DirX := xNone
    else
      DirY := yNone;
    Result := Actions[Kind, DirX, DirY];
  end;
end;

end.
