{------------------------------------------------------------------------------}
{                                                                              }
{  Office Assistant                                                            }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

{$I DELPHIAREA.INC}

unit oaEngine;

interface

uses
  Windows, Classes, Graphics;

type

  TNewFrameEvent = procedure(Sender: TObject; FrameBitmap: TBitmap) of object;

  TNormalBranch = record
    Precent: 0..100;
    Frame: Byte;
  end;

  TExitBranch = record
    Enabled: Boolean;
    Frame: Byte;
  end;

  TAssistantEngine = class(TThread)
  private
    Owner: TPersistent;
    Bitmap, FrameBitmap: TBitmap;
    Delays: array[Byte] of Byte;
    Branches: array[Byte, 1..4] of TNormalBranch;
    ExitBranches: array[Byte] of TExitBranch;
    Sounds: array[Byte] of Word;
    LoadedIndex: Word;
    Returning: Boolean;
    Completed: Boolean;
    FrameCount: Integer;
    CurrentFrame: Byte;
    FActive: Boolean;
    FSpeed: Word;
    FMute: Boolean;
    FOnNewFrame: TNewFrameEvent;
    FOnStop: TNotifyEvent;
    function GetEmpty: Boolean;
    function GetHasSound: Boolean;
    function GetHasExitBranch: Boolean;
    function GetHasInfiniteLoop: Boolean;
    procedure SetActive(Value: Boolean);
    procedure NewFrameAvailable;
    procedure AnimationStopped;
    function FindNextFrame(Frame: Byte; Normal: Boolean): Byte;
  protected
    procedure Execute; override;
  public
    constructor Create(AOwner: TPersistent);
    destructor Destroy; override;
    procedure Clear;
    function Prepare(Index: Word): Boolean;
    function PrepareReturn: Boolean;
    property Empty: Boolean read GetEmpty;
    property HasSound: Boolean read GetHasSound;
    property HasExitBranch: Boolean read GetHasExitBranch;
    property HasInfiniteLoop: Boolean read GetHasInfiniteLoop;
  published
    property Active: Boolean read FActive write SetActive;
    property Mute: Boolean read FMute write FMute;
    property Speed: Word read FSpeed write FSpeed;
    property OnNewFrame: TNewFrameEvent read FOnNewFrame write FOnNewFrame;
    property OnStop: TNotifyEvent read FOnStop write FOnStop;
  end;

procedure ProcessMsg;  // Process non-input messages

implementation

uses
  SysUtils, Messages, MMSystem, Forms, oaActor, oaTypes, oaLZRWKH;

procedure ProcessMsg;
var
  Msg: TMsg;
begin
  PeekMessage(Msg, Application.Handle, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE);
  PeekMessage(Msg, Application.Handle, WM_MOUSEFIRST, WM_MOUSELAST, PM_REMOVE);
end;

{ TAssistantEngine }

constructor TAssistantEngine.Create(AOwner: TPersistent);
begin
  Randomize;
  Owner := AOwner;
  Bitmap := TBitmap.Create;
  FrameBitmap := TBitmap.Create;
  Completed := False;
  LoadedIndex := 0;
  Returning := False;
  FrameCount := 0;
  CurrentFrame := 0;
  FActive := False;
  FSpeed := 100;
  FMute := False;
  FreeOnTerminate := True;
  inherited Create(True);
end;

destructor TAssistantEngine.Destroy;
begin
  Bitmap.Free;
  FrameBitmap.Free;
  inherited Destroy;
end;

procedure TAssistantEngine.Clear;
begin
  Active := False;
  Bitmap.Free;
  Bitmap := TBitmap.Create;
  FrameBitmap.Free;
  FrameBitmap := TBitmap.Create;
  FillChar(Sounds, SizeOf(Sounds), 0);
  FillChar(Delays, SizeOf(Delays), 0);
  FillChar(Branches, SizeOf(Branches), 0);
  FillChar(ExitBranches, SizeOf(ExitBranches), 0);
  FrameCount := 0;
  CurrentFrame := 0;
  Returning := False;
  Completed := False;
  LoadedIndex := 0;
end;

function TAssistantEngine.Prepare(Index: Word): Boolean;
var
  Stream: TMemoryStream;
  ResStream: TResourceStream;
  ScanlineBytes, Row: Integer;
  Bytes: PByteArray;
begin
  Result := False;
  if (Index > 0) and (Index <= TAssistantActor(Owner).AnimationCount) then
  begin
    Active := False;
    if LoadedIndex <> Index then
    begin
      LoadedIndex := Index;
      Stream := TMemoryStream.Create;
      try
        ResStream := TResourceStream.CreateFromID(TAssistantActor(Owner).Instance,
          TAssistantActor(Owner).AnimationInfo[Index].Resource, RT_RCDATA);
        try
          DecompressStream(ResStream, Stream);
        finally
          ResStream.Free;
        end;
        Stream.Position := 0;
        Stream.Read(FrameCount, SizeOf(FrameCount));
        Stream.Read(Sounds, FrameCount * SizeOf(Word));
        Stream.Read(Delays, FrameCount * SizeOf(Byte));
        Stream.Read(Branches, 4 * FrameCount * SizeOf(TNormalBranch));
        Stream.Read(ExitBranches, FrameCount * SizeOf(TExitBranch));
        Bitmap.Width := FrameCount * TAssistantActor(Owner).ScreenWidth;
        if Bitmap.Empty then
        begin
          Bitmap.Height := TAssistantActor(Owner).ScreenHeight;
          Bitmap.PixelFormat := pf8bit;
          Bitmap.Palette := CopyPalette(TAssistantActor(Owner).Palette);
        end;
        Bytes := Bitmap.Scanline[0];
        ScanlineBytes := Integer(Bitmap.Scanline[1]) - Integer(Bytes);
        for Row := 0 to Bitmap.Height-1 do
        begin
          Stream.Read(Bytes^, Bitmap.Width);
          Inc(Integer(Bytes), ScanlineBytes);
        end;
        if FrameBitmap.Empty then
        begin
          FrameBitmap.Width := TAssistantActor(Owner).ScreenWidth;
          FrameBitmap.Height := TAssistantActor(Owner).ScreenHeight;
          FrameBitmap.PixelFormat := pf8bit;
          FrameBitmap.Palette := CopyPalette(TAssistantActor(Owner).Palette);
          FrameBitmap.TransparentColor := TAssistantActor(Owner).TransparentColor;
          FrameBitmap.Transparent := True;
        end;
      finally
        Stream.Free;
      end;
    end;
    Returning := False;
    CurrentFrame := 0;
    Result := True;
  end;
end;

function TAssistantEngine.PrepareReturn: Boolean;
var
  AnimInfo: PAnimationInfo;
begin
  Result := False;
  if (LoadedIndex <> 0) and not Returning then
  begin
    AnimInfo := TAssistantActor(Owner).AnimationInfo[LoadedIndex];
    if ((AnimInfo.ExitType = etExitBranch) or not Completed) and HasExitBranch then
    begin
      Returning := True;
      Result := True;
    end
    else if Completed and (AnimInfo.ExitType = etAnimation) then
      Result := Prepare(AnimInfo^.ReturnIndex);
  end;
end;

function TAssistantEngine.GetHasInfiniteLoop: Boolean;
var
  I, J: Integer;
begin
  Result := False;
  for I := 0 to FrameCount-1 do
    for j := 1 to 4 do
      if (Branches[I,J].Precent = 100) and (Branches[I,J].Frame <= I) then
      begin
        Result := True;
        Exit;
      end;
end;

function TAssistantEngine.GetHasSound: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to FrameCount-1 do
    if Sounds[I] > 0 then
    begin
      Result := True;
      Exit;
    end;
end;

function TAssistantEngine.GetHasExitBranch: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to FrameCount-1 do
    if ExitBranches[I].Enabled then
    begin
      Result := True;
      Exit;
    end;
end;

function TAssistantEngine.GetEmpty: Boolean;
begin
  Result := (FrameCount <= 0);
end;

procedure TAssistantEngine.SetActive(Value: Boolean);
begin
  if Value and (FrameCount = 0) then
    Value := False;
  if Active <> Value then
  begin
    FActive := Value;
    if Active then
    begin
      Completed := False;
      if Suspended then Resume;
    end
    else
    begin
      PlaySound(nil, TAssistantActor(Owner).Instance, SND_RESOURCE or SND_PURGE);
      repeat ProcessMsg until (Terminated or Suspended or Active or Application.Terminated);
    end;
  end;
end;

procedure TAssistantEngine.AnimationStopped;
begin
  if Assigned(FOnStop) then
    FOnStop(Self);
end;

procedure TAssistantEngine.NewFrameAvailable;
var
  SrcRect, DstRect: TRect;
begin
  DstRect := Rect(0, 0, FrameBitmap.Width, FrameBitmap.Height);
  SrcRect := DstRect;
  OffSetRect(SrcRect, CurrentFrame * FrameBitmap.Width, 0);
  FrameBitmap.Canvas.CopyRect(DstRect, Bitmap.Canvas, SrcRect);
  if Assigned(FOnNewFrame) then
    FOnNewFrame(Self, FrameBitmap);
  if (Sounds[CurrentFrame] > 0) and FActive and not FMute then
    PlaySound(MakeIntResource(Sounds[CurrentFrame]), TAssistantActor(Owner).Instance,
      SND_RESOURCE or SND_NOWAIT or SND_ASYNC or SND_NODEFAULT);
end;

function TAssistantEngine.FindNextFrame(Frame: Byte; Normal: Boolean): Byte;
var
  B, P: Byte;
  Branched: Boolean;
begin
  Result := Frame;
  if not Normal then
    if Frame < FrameCount-1 then
      if ExitBranches[Frame].Enabled then
        Result := ExitBranches[Frame].Frame
      else
        Result := Frame + 1
    else
      Completed := True
  else
  begin
    P := 100;
    Branched := False;
    for B := 1 to 4 do
    begin
      if Random(P) < Branches[Frame,B].Precent then
      begin
        Result := Branches[Frame,B].Frame;
        Branched := True;
        Break;
      end
      else
        Dec(P, Branches[Frame,B].Precent);
    end;
    if not Branched then
    begin
      if Frame < FrameCount-1 then
        Result := Frame + 1
      else
        Completed := True;
    end;
  end;
  if (Delays[Result] = 0) and (Result = FrameCount-1) then
    Completed := True;
end;

procedure TAssistantEngine.Execute;
var
  StartTime: DWord;
  DelayTime: Integer;
  NextFrame: Byte;
begin
  while not Terminated do
  begin
    if Active and (Speed > 0) then
    begin
      if Delays[CurrentFrame] > 0 then
      begin
        StartTime := GetTickCount;
        Synchronize(NewFrameAvailable);
        DelayTime := MulDiv(Delays[CurrentFrame], 1000, Speed) -
                     Integer(GetTickCount - StartTime) shr 2;
        if DelayTime >= 0 then SleepEx(DelayTime, True);
      end;
      NextFrame := FindNextFrame(CurrentFrame, not Returning);
      if not Completed and Active then
        CurrentFrame := NextFrame
      else
      begin
        CurrentFrame := FindNextFrame(CurrentFrame, Returning);
        FActive := False;
      end;
    end
    else if not (Active or Terminated) then
    begin
      Synchronize(AnimationStopped);
      if not Active then Suspend;
    end;
  end;
end;

end.

